# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import AssertionEvaluationContextClass
from .....schema_classes import AssertionEvaluationParametersClass
from .....schema_classes import AssertionEvaluationParametersTypeClass
from .....schema_classes import AssertionEvaluationSpecClass
from .....schema_classes import AssertionMonitorClass
from .....schema_classes import AssertionMonitorBootstrapConfigClass
from .....schema_classes import AssertionMonitorBootstrapStatusClass
from .....schema_classes import AssertionMonitorCapabilityClass
from .....schema_classes import AssertionMonitorMetricsCubeBootstrapStateClass
from .....schema_classes import AssertionMonitorMetricsCubeBootstrapStatusClass
from .....schema_classes import AssertionMonitorSettingsClass
from .....schema_classes import AssertionTimeBucketingStrategyClass
from .....schema_classes import AuditLogSpecClass
from .....schema_classes import DataHubOperationSpecClass
from .....schema_classes import DatasetFieldAssertionParametersClass
from .....schema_classes import DatasetFieldAssertionSourceTypeClass
from .....schema_classes import DatasetFreshnessAssertionParametersClass
from .....schema_classes import DatasetFreshnessSourceTypeClass
from .....schema_classes import DatasetSchemaAssertionParametersClass
from .....schema_classes import DatasetSchemaSourceTypeClass
from .....schema_classes import DatasetVolumeAssertionParametersClass
from .....schema_classes import DatasetVolumeSourceTypeClass
from .....schema_classes import EmbeddedAssertionClass
from .....schema_classes import MonitorErrorClass
from .....schema_classes import MonitorErrorTypeClass
from .....schema_classes import MonitorInfoClass
from .....schema_classes import MonitorModeClass
from .....schema_classes import MonitorStateClass
from .....schema_classes import MonitorStatusClass
from .....schema_classes import MonitorSuiteAssertionSettingsClass
from .....schema_classes import MonitorSuiteAssignmentSpecClass
from .....schema_classes import MonitorSuiteInfoClass
from .....schema_classes import MonitorSuiteNotificationSettingsClass
from .....schema_classes import MonitorSuiteTypeClass
from .....schema_classes import MonitorTimeseriesStateClass
from .....schema_classes import MonitorTypeClass


AssertionEvaluationContext = AssertionEvaluationContextClass
AssertionEvaluationParameters = AssertionEvaluationParametersClass
AssertionEvaluationParametersType = AssertionEvaluationParametersTypeClass
AssertionEvaluationSpec = AssertionEvaluationSpecClass
AssertionMonitor = AssertionMonitorClass
AssertionMonitorBootstrapConfig = AssertionMonitorBootstrapConfigClass
AssertionMonitorBootstrapStatus = AssertionMonitorBootstrapStatusClass
AssertionMonitorCapability = AssertionMonitorCapabilityClass
AssertionMonitorMetricsCubeBootstrapState = AssertionMonitorMetricsCubeBootstrapStateClass
AssertionMonitorMetricsCubeBootstrapStatus = AssertionMonitorMetricsCubeBootstrapStatusClass
AssertionMonitorSettings = AssertionMonitorSettingsClass
AssertionTimeBucketingStrategy = AssertionTimeBucketingStrategyClass
AuditLogSpec = AuditLogSpecClass
DataHubOperationSpec = DataHubOperationSpecClass
DatasetFieldAssertionParameters = DatasetFieldAssertionParametersClass
DatasetFieldAssertionSourceType = DatasetFieldAssertionSourceTypeClass
DatasetFreshnessAssertionParameters = DatasetFreshnessAssertionParametersClass
DatasetFreshnessSourceType = DatasetFreshnessSourceTypeClass
DatasetSchemaAssertionParameters = DatasetSchemaAssertionParametersClass
DatasetSchemaSourceType = DatasetSchemaSourceTypeClass
DatasetVolumeAssertionParameters = DatasetVolumeAssertionParametersClass
DatasetVolumeSourceType = DatasetVolumeSourceTypeClass
EmbeddedAssertion = EmbeddedAssertionClass
MonitorError = MonitorErrorClass
MonitorErrorType = MonitorErrorTypeClass
MonitorInfo = MonitorInfoClass
MonitorMode = MonitorModeClass
MonitorState = MonitorStateClass
MonitorStatus = MonitorStatusClass
MonitorSuiteAssertionSettings = MonitorSuiteAssertionSettingsClass
MonitorSuiteAssignmentSpec = MonitorSuiteAssignmentSpecClass
MonitorSuiteInfo = MonitorSuiteInfoClass
MonitorSuiteNotificationSettings = MonitorSuiteNotificationSettingsClass
MonitorSuiteType = MonitorSuiteTypeClass
MonitorTimeseriesState = MonitorTimeseriesStateClass
MonitorType = MonitorTypeClass

# fmt: on



from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, Union

from acryl_datahub_cloud.sdk.assertion.assertion_base import (
    AssertionMode,
    FreshnessAssertion,
)
from acryl_datahub_cloud.sdk.assertion_client.helpers import (
    DEFAULT_CREATED_BY,
    _merge_field,
    _print_experimental_warning,
    retrieve_assertion_and_monitor_by_urn,
)
from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    AssertionIncidentBehaviorInputTypes,
    DetectionMechanismInputTypes,
    TimeWindowSizeInputTypes,
)
from acryl_datahub_cloud.sdk.assertion_input.freshness_assertion_input import (
    FreshnessAssertionScheduleCheckType,
    _FreshnessAssertionInput,
)
from acryl_datahub_cloud.sdk.entities.assertion import Assertion, TagsInputType
from acryl_datahub_cloud.sdk.entities.monitor import Monitor
from acryl_datahub_cloud.sdk.errors import SDKUsageError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn, MonitorUrn

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient

logger = logging.getLogger(__name__)


class FreshnessAssertionClient:
    """Client for managing freshness assertions."""

    def __init__(self, client: "DataHubClient"):
        self.client = client

    def sync_freshness_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        freshness_schedule_check_type: Optional[
            Union[
                str,
                FreshnessAssertionScheduleCheckType,
                models.FreshnessAssertionScheduleTypeClass,
            ]
        ] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        lookback_window: Optional[TimeWindowSizeInputTypes] = None,
    ) -> FreshnessAssertion:
        _print_experimental_warning()
        now_utc = datetime.now(timezone.utc)

        if updated_by is None:
            logger.warning(
                f"updated_by is not set, using {DEFAULT_CREATED_BY} as a placeholder"
            )
            updated_by = DEFAULT_CREATED_BY

        # 1. Retrieve and merge the assertion input with any existing assertion and monitor entities,
        # or build a new assertion input if the assertion does not exist:
        assertion_input = self._retrieve_and_merge_freshness_assertion_and_monitor(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            now_utc=now_utc,
            schedule=schedule,
            freshness_schedule_check_type=freshness_schedule_check_type,
            lookback_window=lookback_window,
        )

        # 2. Upsert the assertion and monitor entities:
        assertion_entity, monitor_entity = (
            assertion_input.to_assertion_and_monitor_entities()
        )
        # If assertion upsert fails, we won't try to upsert the monitor
        self.client.entities.upsert(assertion_entity)
        # TODO: Wrap monitor upsert in a try-except and delete the assertion if monitor upsert fails (once delete is implemented https://linear.app/acryl-data/issue/OBS-1350/add-delete-method-to-entity-clientpy)
        # try:
        self.client.entities.upsert(monitor_entity)
        # except Exception as e:
        #     logger.error(f"Error upserting monitor: {e}")
        #     self.client.entities.delete(assertion_entity)
        #     raise e

        return FreshnessAssertion._from_entities(assertion_entity, monitor_entity)

    def _retrieve_and_merge_freshness_assertion_and_monitor(
        self,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]],
        display_name: Optional[str],
        enabled: Optional[bool],
        detection_mechanism: DetectionMechanismInputTypes,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes],
        tags: Optional[TagsInputType],
        updated_by: Union[str, CorpUserUrn],
        now_utc: datetime,
        schedule: Optional[Union[str, models.CronScheduleClass]],
        freshness_schedule_check_type: Optional[
            Union[str, models.FreshnessAssertionScheduleTypeClass]
        ] = None,
        lookback_window: Optional[TimeWindowSizeInputTypes] = None,
    ) -> _FreshnessAssertionInput:
        # 1. If urn is not provided, build and return a new assertion input directly
        if urn is None:
            logger.info("URN is not set, building a new assertion input")
            return _FreshnessAssertionInput(
                urn=None,
                entity_client=self.client.entities,
                dataset_urn=dataset_urn,
                display_name=display_name,
                enabled=enabled if enabled is not None else True,
                detection_mechanism=detection_mechanism,
                freshness_schedule_check_type=freshness_schedule_check_type,
                lookback_window=lookback_window,
                incident_behavior=incident_behavior,
                tags=tags,
                created_by=updated_by,
                created_at=now_utc,
                updated_by=updated_by,
                updated_at=now_utc,
                schedule=schedule,
            )

        # 2. Build initial assertion input for validation
        assertion_input = _FreshnessAssertionInput(
            urn=urn,
            entity_client=self.client.entities,
            dataset_urn=dataset_urn,
            display_name=display_name,
            detection_mechanism=detection_mechanism,
            freshness_schedule_check_type=freshness_schedule_check_type,
            lookback_window=lookback_window,
            incident_behavior=incident_behavior,
            tags=tags,
            created_by=updated_by,  # This will be overridden by the actual created_by
            created_at=now_utc,  # This will be overridden by the actual created_at
            updated_by=updated_by,
            updated_at=now_utc,
            schedule=schedule,
        )

        # 3. Retrieve any existing assertion and monitor entities:
        maybe_assertion_entity, monitor_urn, maybe_monitor_entity = (
            self._retrieve_assertion_and_monitor(assertion_input)
        )

        # 4.1 If the assertion and monitor entities exist, create an assertion object from them:
        if maybe_assertion_entity and maybe_monitor_entity:
            existing_assertion = FreshnessAssertion._from_entities(
                maybe_assertion_entity, maybe_monitor_entity
            )
        # 4.2 If the assertion exists but the monitor does not, create a placeholder monitor entity to be able to create the assertion:
        elif maybe_assertion_entity and not maybe_monitor_entity:
            monitor_mode = (
                "ACTIVE" if enabled else "INACTIVE" if enabled is not None else "ACTIVE"
            )
            existing_assertion = FreshnessAssertion._from_entities(
                maybe_assertion_entity,
                Monitor(id=monitor_urn, info=("ASSERTION", monitor_mode)),
            )
        # 4.3 If the assertion does not exist, build and return a new assertion input:
        elif not maybe_assertion_entity:
            logger.info(
                f"No existing assertion entity found for assertion urn {urn}, building a new assertion input"
            )
            return _FreshnessAssertionInput(
                urn=urn,
                entity_client=self.client.entities,
                dataset_urn=dataset_urn,
                display_name=display_name,
                enabled=enabled if enabled is not None else True,
                detection_mechanism=detection_mechanism,
                freshness_schedule_check_type=freshness_schedule_check_type,
                lookback_window=lookback_window,
                incident_behavior=incident_behavior,
                tags=tags,
                created_by=updated_by,
                created_at=now_utc,
                updated_by=updated_by,
                updated_at=now_utc,
                schedule=schedule,
            )

        # 5. Check for any issues e.g. different dataset urns
        if (
            existing_assertion
            and hasattr(existing_assertion, "dataset_urn")
            and existing_assertion.dataset_urn != assertion_input.dataset_urn
        ):
            raise SDKUsageError(
                f"Dataset URN mismatch, existing assertion: {existing_assertion.dataset_urn} != new assertion: {dataset_urn}"
            )

        # 6. Merge the existing assertion with the validated input:
        merged_assertion_input = self._merge_freshness_input(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            now_utc=now_utc,
            assertion_input=assertion_input,
            maybe_assertion_entity=maybe_assertion_entity,
            maybe_monitor_entity=maybe_monitor_entity,
            existing_assertion=existing_assertion,
            schedule=schedule,
            freshness_schedule_check_type=freshness_schedule_check_type,
            lookback_window=lookback_window,
        )

        return merged_assertion_input

    def _retrieve_assertion_and_monitor(
        self,
        assertion_input: _FreshnessAssertionInput,
    ) -> tuple[Optional[Assertion], MonitorUrn, Optional[Monitor]]:
        """Retrieve the assertion and monitor entities from the DataHub instance.

        Args:
            assertion_input: The validated input to the function.

        Returns:
            The assertion and monitor entities.
        """
        assert assertion_input.urn is not None, "URN is required"
        return retrieve_assertion_and_monitor_by_urn(
            self.client, assertion_input.urn, assertion_input.dataset_urn
        )

    def _merge_freshness_input(
        self,
        dataset_urn: Union[str, DatasetUrn],
        urn: Union[str, AssertionUrn],
        display_name: Optional[str],
        enabled: Optional[bool],
        detection_mechanism: DetectionMechanismInputTypes,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes],
        tags: Optional[TagsInputType],
        now_utc: datetime,
        assertion_input: _FreshnessAssertionInput,
        maybe_assertion_entity: Optional[Assertion],
        maybe_monitor_entity: Optional[Monitor],
        existing_assertion: FreshnessAssertion,
        schedule: Optional[Union[str, models.CronScheduleClass]],
        freshness_schedule_check_type: Optional[
            Union[str, models.FreshnessAssertionScheduleTypeClass]
        ] = None,
        lookback_window: Optional[TimeWindowSizeInputTypes] = None,
    ) -> _FreshnessAssertionInput:
        """Merge the input with the existing assertion and monitor entities.

        Args:
            dataset_urn: The urn of the dataset to be monitored.
            urn: The urn of the assertion.
            display_name: The display name of the assertion.
            enabled: Whether the assertion is enabled.
            incident_behavior: The incident behavior to be applied to the assertion.
            tags: The tags to be applied to the assertion.
            now_utc: The current UTC time from when the function is called.
            assertion_input: The validated input to the function.
            maybe_assertion_entity: The existing assertion entity from the DataHub instance.
            maybe_monitor_entity: The existing monitor entity from the DataHub instance.
            existing_assertion: The existing assertion from the DataHub instance.
            schedule: The schedule to be applied to the assertion.
            freshness_schedule_check_type: The freshness schedule check type to be applied to the assertion.
            lookback_window: The lookback window to be applied to the assertion.

        Returns:
            The merged assertion input.
        """
        merged_assertion_input = _FreshnessAssertionInput(
            urn=urn,
            entity_client=self.client.entities,
            dataset_urn=dataset_urn,
            display_name=_merge_field(
                display_name,
                "display_name",
                assertion_input,
                existing_assertion,
                maybe_assertion_entity.description if maybe_assertion_entity else None,
            ),
            enabled=_merge_field(
                enabled,
                "enabled",
                assertion_input,
                existing_assertion,
                existing_assertion.mode == AssertionMode.ACTIVE
                if existing_assertion
                else None,
            ),
            schedule=_merge_field(
                schedule,
                "schedule",
                assertion_input,
                existing_assertion,
                existing_assertion.schedule if existing_assertion else None,
            ),
            freshness_schedule_check_type=_merge_field(
                freshness_schedule_check_type,
                "freshness_schedule_check_type",
                assertion_input,
                existing_assertion,
                existing_assertion._freshness_schedule_check_type
                if existing_assertion
                else None,
            ),
            lookback_window=_merge_field(
                lookback_window,
                "lookback_window",
                assertion_input,
                existing_assertion,
                existing_assertion.lookback_window if existing_assertion else None,
            ),
            detection_mechanism=_merge_field(
                detection_mechanism,
                "detection_mechanism",
                assertion_input,
                existing_assertion,
                FreshnessAssertion._get_detection_mechanism(
                    maybe_assertion_entity, maybe_monitor_entity, default=None
                )
                if maybe_assertion_entity and maybe_monitor_entity
                else None,
            ),
            incident_behavior=_merge_field(
                incident_behavior,
                "incident_behavior",
                assertion_input,
                existing_assertion,
                FreshnessAssertion._get_incident_behavior(maybe_assertion_entity)
                if maybe_assertion_entity
                else None,
            ),
            tags=_merge_field(
                tags,
                "tags",
                assertion_input,
                existing_assertion,
                maybe_assertion_entity.tags if maybe_assertion_entity else None,
            ),
            created_by=existing_assertion.created_by
            or DEFAULT_CREATED_BY,  # Override with the existing assertion's created_by or the default created_by if not set
            created_at=existing_assertion.created_at
            or now_utc,  # Override with the existing assertion's created_at or now if not set
            updated_by=assertion_input.updated_by,  # Override with the input's updated_by
            updated_at=assertion_input.updated_at,  # Override with the input's updated_at (now)
        )
        return merged_assertion_input



"""
Schema assertion module.

This module provides the SchemaAssertion class for representing and working with
schema assertions that validate dataset schemas match expected field definitions.
"""

import logging
from datetime import datetime
from typing import Optional, Union

from typing_extensions import Self

from acryl_datahub_cloud.sdk.assertion.assertion_base import (
    AssertionMode,
    _AssertionPublic,
    _HasSchedule,
)
from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    AssertionIncidentBehavior,
    _DetectionMechanismTypes,
)
from acryl_datahub_cloud.sdk.assertion_input.schema_assertion_input import (
    DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY,
    SCHEMA_FIELD_CLASS_TO_TYPE_MAP,
    SchemaAssertionCompatibility,
    SchemaAssertionField,
)
from acryl_datahub_cloud.sdk.entities.assertion import Assertion
from acryl_datahub_cloud.sdk.entities.monitor import Monitor
from acryl_datahub_cloud.sdk.errors import SDKNotYetSupportedError
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn, TagUrn

logger = logging.getLogger(__name__)


class SchemaAssertion(_HasSchedule, _AssertionPublic):
    """
    A class that represents a schema assertion.

    Schema assertions validate that a dataset's schema matches expected
    field definitions with configurable compatibility modes.
    """

    def __init__(
        self,
        *,
        urn: AssertionUrn,
        dataset_urn: DatasetUrn,
        display_name: str,
        mode: AssertionMode,
        schedule: models.CronScheduleClass,
        compatibility: SchemaAssertionCompatibility,
        fields: list[SchemaAssertionField],
        tags: list[TagUrn],
        incident_behavior: list[AssertionIncidentBehavior],
        created_by: Optional[CorpUserUrn] = None,
        created_at: Union[datetime, None] = None,
        updated_by: Optional[CorpUserUrn] = None,
        updated_at: Optional[datetime] = None,
    ):
        """
        Initialize a schema assertion.

        Note: Values can be accessed, but not set on the assertion object.
        To update an assertion, use the `sync_schema_assertion` method.

        Args:
            urn: The URN of the assertion.
            dataset_urn: The URN of the dataset that the assertion validates.
            display_name: The display name of the assertion.
            mode: The mode of the assertion (active, inactive).
            schedule: The evaluation schedule of the assertion.
            compatibility: The compatibility mode (EXACT_MATCH, SUPERSET, SUBSET).
            fields: The expected schema fields to validate.
            tags: The tags applied to the assertion.
            incident_behavior: Actions to take on assertion pass/fail.
            created_by: The user who created the assertion.
            created_at: The timestamp when the assertion was created.
            updated_by: The user who last updated the assertion.
            updated_at: The timestamp when the assertion was last updated.
        """
        _HasSchedule.__init__(self, schedule=schedule)
        _AssertionPublic.__init__(
            self,
            urn=urn,
            dataset_urn=dataset_urn,
            display_name=display_name,
            mode=mode,
            incident_behavior=incident_behavior,
            detection_mechanism=None,  # Schema assertions don't use detection mechanism
            created_by=created_by,
            created_at=created_at,
            updated_by=updated_by,
            updated_at=updated_at,
            tags=tags,
        )
        self._compatibility = compatibility
        self._fields = fields

    @property
    def compatibility(self) -> SchemaAssertionCompatibility:
        """The compatibility mode for the schema assertion."""
        return self._compatibility

    @property
    def fields(self) -> list[SchemaAssertionField]:
        """The expected schema fields to validate."""
        return self._fields

    @staticmethod
    def _get_detection_mechanism(
        assertion: Assertion,
        monitor: Monitor,
        default: Optional[_DetectionMechanismTypes] = None,
    ) -> Optional[_DetectionMechanismTypes]:
        """
        Schema assertions don't have a detection mechanism.
        They always use DATAHUB_SCHEMA as the source type.
        """
        return None

    @staticmethod
    def _get_compatibility(assertion: Assertion) -> SchemaAssertionCompatibility:
        """Extract compatibility from the assertion info."""
        if assertion.info is None:
            logger.warning(
                f"Assertion {assertion.urn} does not have info, "
                f"defaulting to {DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY}"
            )
            return DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY

        if not isinstance(assertion.info, models.SchemaAssertionInfoClass):
            raise SDKNotYetSupportedError(
                f"Assertion {assertion.urn} is not a schema assertion"
            )

        compatibility_value = assertion.info.compatibility
        if compatibility_value is None:
            return DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY

        # Handle enum value extraction
        try:
            if hasattr(compatibility_value, "name"):
                # It's an enum-like object
                return SchemaAssertionCompatibility(compatibility_value.name)
            elif isinstance(compatibility_value, str):
                return SchemaAssertionCompatibility(compatibility_value.upper())
            else:
                logger.warning(
                    f"Unknown compatibility value type {type(compatibility_value)}, "
                    f"defaulting to {DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY}"
                )
                return DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY
        except ValueError:
            logger.warning(
                f"Unknown compatibility value '{compatibility_value}', "
                f"defaulting to {DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY}"
            )
            return DEFAULT_SCHEMA_ASSERTION_COMPATIBILITY

    @staticmethod
    def _get_fields(assertion: Assertion) -> list[SchemaAssertionField]:
        """Extract schema fields from the assertion info."""
        if assertion.info is None:
            logger.warning(
                f"Assertion {assertion.urn} does not have info, returning empty fields"
            )
            return []

        if not isinstance(assertion.info, models.SchemaAssertionInfoClass):
            raise SDKNotYetSupportedError(
                f"Assertion {assertion.urn} is not a schema assertion"
            )

        schema = assertion.info.schema
        if schema is None:
            logger.warning(
                f"Assertion {assertion.urn} does not have schema metadata, "
                "returning empty fields"
            )
            return []

        fields = []
        failed_fields = []
        for schema_field in schema.fields or []:
            field = SchemaAssertion._parse_schema_field(schema_field)
            if field is not None:
                fields.append(field)
            else:
                failed_fields.append(schema_field.fieldPath)

        if failed_fields:
            logger.warning(
                f"Failed to parse {len(failed_fields)} field(s) from assertion {assertion.urn}: "
                f"{', '.join(failed_fields)}. These fields will be excluded from the assertion object. "
                f"This may indicate unsupported field types or corrupted data."
            )

        return fields

    @staticmethod
    def _parse_schema_field(
        schema_field: models.SchemaFieldClass,
    ) -> Optional[SchemaAssertionField]:
        """Parse a SchemaFieldClass into a SchemaAssertionField."""
        # Explicit attribute checks to fail fast on missing required attributes
        if not hasattr(schema_field, "fieldPath") or not hasattr(schema_field, "type"):
            logger.warning(
                "Schema field missing required attributes: fieldPath or type, skipping"
            )
            return None

        # Extract the type from the SchemaFieldDataTypeClass
        field_data_type = schema_field.type
        if field_data_type is None or field_data_type.type is None:
            logger.warning(
                f"Schema field {schema_field.fieldPath} has no type, skipping"
            )
            return None

        # Get the actual type class instance
        type_instance = field_data_type.type
        type_class = type(type_instance)

        # Look up the SchemaFieldDataType enum value
        field_type = SCHEMA_FIELD_CLASS_TO_TYPE_MAP.get(type_class)
        if field_type is None:
            logger.warning(
                f"Unknown schema field type {type_class.__name__} "
                f"for field {schema_field.fieldPath}, skipping"
            )
            return None

        return SchemaAssertionField(
            path=schema_field.fieldPath,
            type=field_type,
            native_type=getattr(schema_field, "nativeDataType", None),
        )

    @classmethod
    def _from_entities(cls, assertion: Assertion, monitor: Monitor) -> Self:
        """
        Create a SchemaAssertion from the assertion and monitor entities.

        Args:
            assertion: The assertion entity from the backend.
            monitor: The monitor entity from the backend.

        Returns:
            A SchemaAssertion instance populated from the entities.
        """
        return cls(
            urn=assertion.urn,
            dataset_urn=assertion.dataset,
            display_name=assertion.description or "",
            mode=cls._get_mode(monitor),
            schedule=cls._get_schedule(monitor),
            compatibility=cls._get_compatibility(assertion),
            fields=cls._get_fields(assertion),
            incident_behavior=cls._get_incident_behavior(assertion),
            created_by=cls._get_created_by(assertion),
            created_at=cls._get_created_at(assertion),
            updated_by=cls._get_updated_by(assertion),
            updated_at=cls._get_updated_at(assertion),
            tags=cls._get_tags(assertion),
        )



from __future__ import annotations

import functools
import logging
import time
from typing import TYPE_CHECKING, Optional, Union

from acryl_datahub_cloud.sdk.assertion.assertion_base import (
    FreshnessAssertion,
    SmartFreshnessAssertion,
    SmartVolumeAssertion,
    SqlAssertion,
    VolumeAssertion,
)
from acryl_datahub_cloud.sdk.assertion.column_metric_assertion import (
    ColumnMetricAssertion,
)
from acryl_datahub_cloud.sdk.assertion.column_value_assertion import (
    ColumnValueAssertion,
)
from acryl_datahub_cloud.sdk.assertion.schema_assertion import SchemaAssertion
from acryl_datahub_cloud.sdk.assertion.smart_column_metric_assertion import (
    SmartColumnMetricAssertion,
)
from acryl_datahub_cloud.sdk.assertion.smart_sql_assertion import SmartSqlAssertion
from acryl_datahub_cloud.sdk.assertion_client.column_metric import (
    ColumnMetricAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.column_value import (
    ColumnValueAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.freshness import (
    FreshnessAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.helpers import (
    _print_experimental_warning,
)
from acryl_datahub_cloud.sdk.assertion_client.schema import (
    SchemaAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.smart_column_metric import (
    SmartColumnMetricAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.smart_freshness import (
    SmartFreshnessAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.smart_sql import (
    SmartSqlAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.smart_volume import (
    SmartVolumeAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.sql import (
    SqlAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_client.volume import (
    VolumeAssertionClient,
)
from acryl_datahub_cloud.sdk.assertion_input.assertion_input import (
    AssertionIncidentBehaviorInputTypes,
    BackfillConfigInputTypes,
    DetectionMechanismInputTypes,
    ExclusionWindowInputTypes,
    InferenceSensitivity,
    TimeBucketingStrategyInputTypes,
    TimeWindowSizeInputTypes,
)
from acryl_datahub_cloud.sdk.assertion_input.column_metric_assertion_input import (
    ColumnMetricAssertionParameters,
)
from acryl_datahub_cloud.sdk.assertion_input.column_metric_constants import (
    MetricInputType,
    OperatorInputType,
)
from acryl_datahub_cloud.sdk.assertion_input.column_value_assertion_input import (
    ColumnValueAssertionParameters,
    FailThresholdInputType,
    FieldTransformInputType,
)
from acryl_datahub_cloud.sdk.assertion_input.freshness_assertion_input import (
    FreshnessAssertionScheduleCheckType,
)
from acryl_datahub_cloud.sdk.assertion_input.schema_assertion_input import (
    SchemaAssertionCompatibility,
    SchemaAssertionFieldsInputType,
)
from acryl_datahub_cloud.sdk.assertion_input.sql_assertion_input import (
    SqlAssertionCondition,
)
from acryl_datahub_cloud.sdk.assertion_input.volume_assertion_input import (
    VolumeAssertionCondition,
    VolumeAssertionDefinitionParameters,
)
from acryl_datahub_cloud.sdk.entities.assertion import TagsInputType
from acryl_datahub_cloud.sdk.errors import SDKUsageError
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata import schema_classes as models
from datahub.metadata.urns import AssertionUrn, CorpUserUrn, DatasetUrn

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient

logger = logging.getLogger(__name__)

# TODO: Replace __datahub_system with the actual datahub system user https://linear.app/acryl-data/issue/OBS-1351/auditstamp-actor-hydration-pattern-for-sdk-calls
DEFAULT_CREATED_BY = CorpUserUrn.from_string("urn:li:corpuser:__datahub_system")


class AssertionsClient:
    def __init__(self, client: "DataHubClient"):
        self.client = client
        self._freshness_client = FreshnessAssertionClient(client)
        self._volume_client = VolumeAssertionClient(client)
        self._sql_client = SqlAssertionClient(client)
        self._smart_freshness_client = SmartFreshnessAssertionClient(client)
        self._smart_volume_client = SmartVolumeAssertionClient(client)
        self._smart_sql_client = SmartSqlAssertionClient(client)
        self._smart_column_metric_client = SmartColumnMetricAssertionClient(client)
        self._column_metric_client = ColumnMetricAssertionClient(client)
        self._schema_client = SchemaAssertionClient(client)
        self._column_value_client = ColumnValueAssertionClient(client)
        # Create a cached version of the existence check with TTL using time bucketing
        # The time_bucket parameter is used only as a cache key to invalidate entries
        # every 60 seconds - it's not used in the function body itself
        self._cached_exists = functools.lru_cache(maxsize=128)(
            lambda urn, _time_bucket: self._graph.exists(urn)
        )
        _print_experimental_warning()

    @property
    def _graph(self) -> DataHubGraph:
        """Access to the underlying DataHubGraph client."""
        return self.client._graph

    def _check_dataset_exists(
        self, dataset_urn: Union[str, DatasetUrn], skip_check: bool
    ) -> None:
        """Verify the dataset exists in DataHub.

        Uses a short-lived cache (60s TTL) via functools.lru_cache to avoid
        redundant network calls when checking multiple assertions for the same dataset.

        Args:
            dataset_urn: The URN of the dataset to check.
            skip_check: If True, skip the existence check.

        Raises:
            SDKUsageError: If the dataset does not exist and skip_check is False.
        """
        if not skip_check:
            dataset_urn_str = str(dataset_urn)
            # Use time bucketing for TTL: bucket time into 60-second intervals
            time_bucket = int(time.time() // 60)
            exists = self._cached_exists(dataset_urn_str, time_bucket)

            if not exists:
                raise SDKUsageError(f"Dataset {dataset_urn_str} does not exist")

    def sync_smart_freshness_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        sensitivity: Optional[Union[str, InferenceSensitivity]] = None,
        exclusion_windows: Optional[ExclusionWindowInputTypes] = None,
        training_data_lookback_days: Optional[int] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        skip_dataset_exists_check: bool = False,
    ) -> SmartFreshnessAssertion:
        """Upsert and merge a smart freshness assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *")
            - Update case: Preserves existing schedule from backend (not modifiable)

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Information schema is recommended. Valid values are:
                - "information_schema" or DetectionMechanism.INFORMATION_SCHEMA
                - "audit_log" or DetectionMechanism.AUDIT_LOG
                - {"type": "last_modified_column", "column_name": "last_modified", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.LAST_MODIFIED_COLUMN(column_name='last_modified', additional_filter='last_modified > 2021-01-01')
                - "datahub_operation" or DetectionMechanism.DATAHUB_OPERATION
            sensitivity (Optional[Union[str, InferenceSensitivity]]): The sensitivity to be applied to the assertion. Valid values are: "low", "medium", "high".
            exclusion_windows (Optional[ExclusionWindowInputTypes]): The exclusion windows to be applied to the assertion. Only fixed range exclusion windows are supported. Valid values are:
                - {"start": "2025-01-01T00:00:00", "end": "2025-01-02T00:00:00"} (using ISO strings)
                - {"start": datetime(2025, 1, 1, 0, 0, 0), "end": datetime(2025, 1, 2, 0, 0, 0)} (using datetime objects)
                - FixedRangeExclusionWindow(start=datetime(2025, 1, 1, 0, 0, 0), end=datetime(2025, 1, 2, 0, 0, 0)) (using typed object)
                - A list of any of the above formats
            training_data_lookback_days (Optional[int]): The training data lookback days to be applied to the assertion as an integer.
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass" or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SmartFreshnessAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._smart_freshness_client.sync_smart_freshness_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
        )

    def sync_smart_volume_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        sensitivity: Optional[Union[str, InferenceSensitivity]] = None,
        exclusion_windows: Optional[ExclusionWindowInputTypes] = None,
        training_data_lookback_days: Optional[int] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        time_bucketing_strategy: TimeBucketingStrategyInputTypes = None,
        backfill_config: BackfillConfigInputTypes = None,
        skip_dataset_exists_check: bool = False,
    ) -> SmartVolumeAssertion:
        """Upsert and merge a smart volume assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Schedule is updated if provided, otherwise existing schedule is preserved.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Information schema is recommended. Valid values are:
                - "information_schema" or DetectionMechanism.INFORMATION_SCHEMA
                - {"type": "query", "additional_filter": "value > 1000"} or DetectionMechanism.QUERY(additional_filter='value > 1000')
                - "dataset_profile" or DetectionMechanism.DATASET_PROFILE
            sensitivity (Optional[Union[str, InferenceSensitivity]]): The sensitivity to be applied to the assertion. Valid values are: "low", "medium", "high".
            exclusion_windows (Optional[ExclusionWindowInputTypes]): The exclusion windows to be applied to the assertion. Only fixed range exclusion windows are supported. Valid values are:
                - {"start": "2025-01-01T00:00:00", "end": "2025-01-02T00:00:00"} (using ISO strings)
                - {"start": datetime(2025, 1, 1, 0, 0, 0), "end": datetime(2025, 1, 2, 0, 0, 0)} (using datetime objects)
                - FixedRangeExclusionWindow(start=datetime(2025, 1, 1, 0, 0, 0), end=datetime(2025, 1, 2, 0, 0, 0)) (using typed object)
                - A list of any of the above formats
            training_data_lookback_days (Optional[int]): The training data lookback days to be applied to the assertion as an integer.
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            time_bucketing_strategy (Optional[TimeBucketingStrategyInputTypes]): Optional time bucketing strategy for evaluating this assertion over time-based data partitions. Accepts a TimeBucketingStrategy Pydantic model, a dict, or a raw AssertionTimeBucketingStrategyClass. When updating, preserves the existing value if not provided. Bucket interval unit must be DAY or WEEK with multiple=1.
            backfill_config (Optional[BackfillConfigInputTypes]): Optional backfill configuration for bootstrapping historical metrics. Specifies the earliest date (as epoch ms or datetime) to backfill from. Accepts a BackfillConfig Pydantic model, a dict with ``backfill_start_date_ms``, or a raw AssertionMonitorBootstrapConfigClass. When updating, preserves the existing value if not provided. Backend enforces max lookback of 365 days (DAY bucketing) or 156 weeks (WEEK bucketing).
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SmartVolumeAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._smart_volume_client.sync_smart_volume_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
            time_bucketing_strategy=time_bucketing_strategy,
            backfill_config=backfill_config,
        )

    def sync_column_metric_assertion(  # TODO: Refactor
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        column_name: Optional[str] = None,
        metric_type: Optional[MetricInputType] = None,
        operator: Optional[OperatorInputType] = None,
        criteria_parameters: Optional[ColumnMetricAssertionParameters] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        time_bucketing_strategy: TimeBucketingStrategyInputTypes = None,
        skip_dataset_exists_check: bool = False,
    ) -> ColumnMetricAssertion:
        """Upsert and merge a column metric assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated.

        Existing assertion fields will be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Examples:
            # Using enum values (recommended for type safety)
            from acryl_datahub_cloud.sdk.assertion_input.column_metric_constants import MetricType, OperatorType
            client.sync_column_metric_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="user_id",
                metric_type=MetricType.NULL_COUNT,
                operator=OperatorType.GREATER_THAN,
                criteria_parameters=10
            )

            # Using case-insensitive strings (more flexible)
            client.sync_column_metric_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="price",
                metric_type="mean",
                operator="between",
                criteria_parameters=(100.0, 500.0)
            )

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            column_name (Optional[str]): The name of the column to be monitored. Required for creation, optional for updates.
            metric_type (Optional[MetricInputType]): The type of the metric to be monitored. Required for creation, optional for updates. Valid values are:
                - Using MetricType enum: MetricType.NULL_COUNT, MetricType.NULL_PERCENTAGE, MetricType.UNIQUE_COUNT,
                  MetricType.UNIQUE_PERCENTAGE, MetricType.MAX_LENGTH, MetricType.MIN_LENGTH, MetricType.EMPTY_COUNT,
                  MetricType.EMPTY_PERCENTAGE, MetricType.MIN, MetricType.MAX, MetricType.MEAN, MetricType.MEDIAN,
                  MetricType.STDDEV, MetricType.NEGATIVE_COUNT, MetricType.NEGATIVE_PERCENTAGE, MetricType.ZERO_COUNT,
                  MetricType.ZERO_PERCENTAGE
                - Using case-insensitive strings: "null_count", "MEAN", "Max_Length", etc.
                - Using models enum: models.FieldMetricTypeClass.NULL_COUNT, etc. (import with: from datahub.metadata import schema_classes as models)
            operator (Optional[OperatorInputType]): The operator to be used for the assertion. Required for creation, optional for updates. Valid values are:
                - Using OperatorType enum: OperatorType.EQUAL_TO, OperatorType.NOT_EQUAL_TO, OperatorType.GREATER_THAN,
                  OperatorType.GREATER_THAN_OR_EQUAL_TO, OperatorType.LESS_THAN, OperatorType.LESS_THAN_OR_EQUAL_TO,
                  OperatorType.BETWEEN, OperatorType.IN, OperatorType.NOT_IN, OperatorType.NULL, OperatorType.NOT_NULL,
                  OperatorType.IS_TRUE, OperatorType.IS_FALSE, OperatorType.CONTAIN, OperatorType.END_WITH,
                  OperatorType.START_WITH, OperatorType.REGEX_MATCH
                - Using case-insensitive strings: "equal_to", "not_equal_to", "greater_than", "greater_than_or_equal_to",
                  "less_than", "less_than_or_equal_to", "between", "in", "not_in", "null", "not_null", "is_true",
                  "is_false", "contain", "end_with", "start_with", "regex_match"
                - Using models enum: models.AssertionStdOperatorClass.EQUAL_TO, models.AssertionStdOperatorClass.GREATER_THAN, etc.
            criteria_parameters (Optional[ColumnMetricAssertionParameters]): The criteria parameters for the assertion. Required for creation (except for operators that don't need parameters), optional for updates.
                - Single value operators (EQUAL_TO, NOT_EQUAL_TO, GREATER_THAN, GREATER_THAN_OR_EQUAL_TO, LESS_THAN, LESS_THAN_OR_EQUAL_TO, CONTAIN, END_WITH, START_WITH, REGEX_MATCH): pass a single number or string
                - Range operators (BETWEEN): pass a tuple of two numbers (min_value, max_value)
                - List operators (IN, NOT_IN): pass a list of values
                - No parameter operators (NULL, NOT_NULL, IS_TRUE, IS_FALSE): pass None or omit this parameter
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Valid values are (additional_filter is optional):
                - "all_rows_query_datahub_dataset_profile" or DetectionMechanism.ALL_ROWS_QUERY_DATAHUB_DATASET_PROFILE
                - "all_rows_query" or DetectionMechanism.ALL_ROWS_QUERY(), or with additional_filter: {"type": "all_rows_query", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.ALL_ROWS_QUERY(additional_filter='last_modified > 2021-01-01')
                - {"type": "changed_rows_query", "column_name": "last_modified", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.CHANGED_ROWS_QUERY(column_name='last_modified', additional_filter='last_modified > 2021-01-01')
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            time_bucketing_strategy (Optional[TimeBucketingStrategyInputTypes]): Optional time bucketing strategy for evaluating this assertion over time-based data partitions. Accepts a TimeBucketingStrategy Pydantic model, a dict, or a raw AssertionTimeBucketingStrategyClass. When updating, preserves the existing value if not provided. Bucket interval unit must be DAY or WEEK with multiple=1. Detection mechanism must be ALL_ROWS_QUERY when bucketing is enabled.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            ColumnMetricAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._column_metric_client.sync_column_metric_assertion(
            dataset_urn=dataset_urn,
            column_name=column_name,
            metric_type=metric_type,
            operator=operator,
            criteria_parameters=criteria_parameters,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
            time_bucketing_strategy=time_bucketing_strategy,
        )

    def sync_column_value_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        column_name: Optional[str] = None,
        operator: Optional[OperatorInputType] = None,
        criteria_parameters: Optional[ColumnValueAssertionParameters] = None,
        transform: Optional[FieldTransformInputType] = None,
        fail_threshold_type: Optional[FailThresholdInputType] = None,
        fail_threshold_value: Optional[int] = None,
        exclude_nulls: Optional[bool] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        skip_dataset_exists_check: bool = False,
    ) -> ColumnValueAssertion:
        """Upsert and merge a column value assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated.

        Existing assertion fields will be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Column value assertions validate individual row values in a column against
        semantic constraints (e.g., "all values must match pattern X" or "no NULL values allowed").
        This differs from column metric assertions which validate aggregated metrics.

        Schedule behavior:
            - Create case: Uses default schedule of every 6 hours or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Examples:
            # Simple email regex validation
            client.assertions.sync_column_value_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="email",
                operator="regex_match",
                criteria_parameters=r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\\.[a-zA-Z]{2,}$"
            )

            # Validate quantity is positive with 5% failure tolerance
            client.assertions.sync_column_value_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="quantity",
                operator="greater_than",
                criteria_parameters=0,
                fail_threshold_type="percentage",
                fail_threshold_value=5,
                exclude_nulls=True
            )

            # Range validation with transform
            client.assertions.sync_column_value_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="description",
                operator="between",
                criteria_parameters=(10, 500),
                transform="length",  # Only for STRING columns
                schedule="0 */6 * * *"
            )

            # SQL expression for validating against a reference table
            from acryl_datahub_cloud.sdk import SqlExpression
            client.assertions.sync_column_value_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="customer_id",
                operator="IN",
                criteria_parameters=SqlExpression("SELECT id FROM valid_customers WHERE active = true"),
            )

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            column_name (Optional[str]): The name of the column to validate. Required for creation, optional for updates.
            operator (Optional[OperatorInputType]): The operator to use for validation. Required for creation, optional for updates. Valid values are:
                - Using OperatorType enum: OperatorType.NOT_NULL, OperatorType.EQUAL_TO, OperatorType.GREATER_THAN,
                  OperatorType.LESS_THAN, OperatorType.BETWEEN, OperatorType.IN, OperatorType.REGEX_MATCH, etc.
                - Using case-insensitive strings: "not_null", "equal_to", "greater_than", "less_than", "between",
                  "in", "regex_match", etc.
                - Using models enum: models.AssertionStdOperatorClass.NOT_NULL, etc.
            criteria_parameters (Optional[ColumnValueAssertionParameters]): The criteria parameters for the operator.
                - Single value operators (EQUAL_TO, GREATER_THAN, LESS_THAN, REGEX_MATCH, etc.): pass a single number or string
                - Range operators (BETWEEN): pass a tuple of two values (min_value, max_value)
                - List operators (IN, NOT_IN): pass a list of values
                - No parameter operators (NOT_NULL, NULL): pass None or omit this parameter
                - SQL expression: pass SqlExpression("SELECT ...") to validate against a SQL query result
            transform (Optional[FieldTransformInputType]): Optional transform to apply to field values before evaluation.
                Currently only "length" or "LENGTH" is supported, and only for STRING columns.
            fail_threshold_type (Optional[FailThresholdInputType]): The type of failure threshold. Valid values are:
                - "count" or "COUNT": Absolute number of failing rows
                - "percentage" or "PERCENTAGE": Percentage of failing rows
                If not provided, defaults to "count" for new assertions.
            fail_threshold_value (Optional[int]): The failure threshold value. For COUNT type, this is the maximum
                number of rows allowed to fail. For PERCENTAGE type, this is the maximum percentage (0-100) allowed to fail.
                Defaults to 0 (no failures allowed) if not provided for new assertions.
            exclude_nulls (Optional[bool]): Whether to exclude null values when evaluating the assertion.
                Defaults to True if not provided for new assertions.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, defaults to True for new assertions.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Valid values are (additional_filter is optional):
                - "all_rows_query_datahub_dataset_profile" or DetectionMechanism.ALL_ROWS_QUERY_DATAHUB_DATASET_PROFILE
                - "all_rows_query" or DetectionMechanism.ALL_ROWS_QUERY(), or with additional_filter: {"type": "all_rows_query", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.ALL_ROWS_QUERY(additional_filter='last_modified > 2021-01-01')
                - {"type": "changed_rows_query", "column_name": "last_modified", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.CHANGED_ROWS_QUERY(column_name='last_modified', additional_filter='last_modified > 2021-01-01')
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default schedule of every 6 hours will be used. The format is a cron expression, e.g. "0 * * * *" for every hour using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            ColumnValueAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._column_value_client.sync_column_value_assertion(
            dataset_urn=dataset_urn,
            column_name=column_name,
            operator=operator,
            criteria_parameters=criteria_parameters,
            transform=transform,
            fail_threshold_type=fail_threshold_type,
            fail_threshold_value=fail_threshold_value,
            exclude_nulls=exclude_nulls,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
        )

    def sync_smart_column_metric_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        column_name: Optional[str] = None,
        metric_type: Optional[MetricInputType] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        sensitivity: Optional[Union[str, InferenceSensitivity]] = None,
        exclusion_windows: Optional[ExclusionWindowInputTypes] = None,
        training_data_lookback_days: Optional[int] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        time_bucketing_strategy: TimeBucketingStrategyInputTypes = None,
        backfill_config: BackfillConfigInputTypes = None,
        skip_dataset_exists_check: bool = False,
    ) -> SmartColumnMetricAssertion:
        """Upsert and merge a smart column metric assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated.

        Existing assertion fields will be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Examples:
            # Using enum values (recommended for type safety)
            client.sync_smart_column_metric_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="user_id",
                metric_type=MetricType.NULL_COUNT
            )

            # Using case-insensitive strings (more flexible)
            client.sync_smart_column_metric_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,database.schema.table,PROD)",
                column_name="price",
                metric_type="mean"
            )

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            column_name (Optional[str]): The name of the column to be monitored. Required for creation, optional for updates.
            metric_type (Optional[MetricInputType]): The type of the metric to be monitored. Required for creation, optional for updates.
                Only count-based metrics are allowed for smart assertions:
                - Using MetricType enum: MetricType.NULL_COUNT, MetricType.UNIQUE_COUNT, MetricType.EMPTY_COUNT,
                  MetricType.NEGATIVE_COUNT, MetricType.ZERO_COUNT
                - Using case-insensitive strings: "null_count", "unique_count", "empty_count", "negative_count", "zero_count"
                - Using models enum: models.FieldMetricTypeClass.NULL_COUNT, etc. (import with: from datahub.metadata import schema_classes as models)
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Valid values are (additional_filter is optional):
                - "all_rows_query_datahub_dataset_profile" or DetectionMechanism.ALL_ROWS_QUERY_DATAHUB_DATASET_PROFILE
                - "all_rows_query" or DetectionMechanism.ALL_ROWS_QUERY(), or with additional_filter: {"type": "all_rows_query", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.ALL_ROWS_QUERY(additional_filter='last_modified > 2021-01-01')
                - {"type": "changed_rows_query", "column_name": "last_modified", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.CHANGED_ROWS_QUERY(column_name='last_modified', additional_filter='last_modified > 2021-01-01')
            sensitivity (Optional[Union[str, InferenceSensitivity]]): The sensitivity to be applied to the assertion. Valid values are: "low", "medium", "high".
            exclusion_windows (Optional[ExclusionWindowInputTypes]): The exclusion windows to be applied to the assertion. Only fixed range exclusion windows are supported.
            training_data_lookback_days (Optional[int]): The training data lookback days to be applied to the assertion as an integer.
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            time_bucketing_strategy (Optional[TimeBucketingStrategyInputTypes]): Optional time bucketing strategy for evaluating this assertion over time-based data partitions. Accepts a TimeBucketingStrategy Pydantic model, a dict, or a raw AssertionTimeBucketingStrategyClass. When updating, preserves the existing value if not provided. Bucket interval unit must be DAY or WEEK with multiple=1. Detection mechanism must be ALL_ROWS_QUERY when bucketing is enabled.
            backfill_config (Optional[BackfillConfigInputTypes]): Optional backfill configuration for bootstrapping historical metrics. Requires time_bucketing_strategy to also be set. Specifies the earliest date (as epoch ms or datetime) to backfill from. Accepts a BackfillConfig Pydantic model, a dict with ``backfill_start_date_ms``, or a raw AssertionMonitorBootstrapConfigClass. When updating, preserves the existing value if not provided. Backend enforces max lookback of 365 days (DAY bucketing) or 156 weeks (WEEK bucketing).
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SmartColumnMetricAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._smart_column_metric_client.sync_smart_column_metric_assertion(
            dataset_urn=dataset_urn,
            column_name=column_name,
            metric_type=metric_type,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
            time_bucketing_strategy=time_bucketing_strategy,
            backfill_config=backfill_config,
        )

    def sync_freshness_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        freshness_schedule_check_type: Optional[
            Union[
                str,
                FreshnessAssertionScheduleCheckType,
                models.FreshnessAssertionScheduleTypeClass,
            ]
        ] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        lookback_window: Optional[TimeWindowSizeInputTypes] = None,
        skip_dataset_exists_check: bool = False,
    ) -> FreshnessAssertion:
        """Upsert and merge a freshness assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Information schema is recommended. Valid values are:
                - "information_schema" or DetectionMechanism.INFORMATION_SCHEMA
                - "audit_log" or DetectionMechanism.AUDIT_LOG
                - {"type": "last_modified_column", "column_name": "last_modified", "additional_filter": "last_modified > '2021-01-01'"} or DetectionMechanism.LAST_MODIFIED_COLUMN(column_name='last_modified', additional_filter='last_modified > 2021-01-01')
                - {"type": "high_watermark_column", "column_name": "id", "additional_filter": "id > 1000"} or DetectionMechanism.HIGH_WATERMARK_COLUMN(column_name='id', additional_filter='id > 1000')
                - "datahub_operation" or DetectionMechanism.DATAHUB_OPERATION
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            freshness_schedule_check_type (Optional[Union[str, FreshnessAssertionScheduleCheckType, models.FreshnessAssertionScheduleTypeClass]]): The freshness schedule check type to be applied to the assertion. Valid values are: "since_the_last_check", "fixed_interval".
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            lookback_window (Optional[TimeWindowSizeInputTypes]): The lookback window to be applied to the assertion. Valid values are:
                - TimeWindowSize(unit=CalendarInterval.MINUTE, multiple=10) for 10 minutes
                - TimeWindowSize(unit=CalendarInterval.HOUR, multiple=2) for 2 hours
                - TimeWindowSize(unit=CalendarInterval.DAY, multiple=1) for 1 day
                - {"unit": "MINUTE", "multiple": 30} for 30 minutes (using dict)
                - {"unit": "HOUR", "multiple": 6} for 6 hours (using dict)
                - {"unit": "DAY", "multiple": 7} for 7 days (using dict)
                Valid values for CalendarInterval are: "MINUTE", "HOUR", "DAY" and for multiple, the integer number of units.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            FreshnessAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._freshness_client.sync_freshness_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            freshness_schedule_check_type=freshness_schedule_check_type,
            schedule=schedule,
            lookback_window=lookback_window,
        )

    def sync_volume_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        detection_mechanism: DetectionMechanismInputTypes = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        criteria_condition: Optional[Union[str, VolumeAssertionCondition]] = None,
        criteria_parameters: Optional[VolumeAssertionDefinitionParameters] = None,
        time_bucketing_strategy: TimeBucketingStrategyInputTypes = None,
        skip_dataset_exists_check: bool = False,
    ) -> VolumeAssertion:
        """Upsert and merge a volume assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            detection_mechanism (DetectionMechanismInputTypes): The detection mechanism to be used for the assertion. Information schema is recommended. Valid values are (additional_filter is optional):
                - "information_schema" or DetectionMechanism.INFORMATION_SCHEMA
                - {"type": "query", "additional_filter": "value > 1000"} or DetectionMechanism.QUERY(additional_filter='value > 1000')
                - "dataset_profile" or DetectionMechanism.DATASET_PROFILE
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            criteria_condition (Optional[Union[str, VolumeAssertionCondition]]): Optional condition for the volume assertion. Valid values are:
                - "ROW_COUNT_IS_LESS_THAN_OR_EQUAL_TO" -> The row count is less than or equal to the threshold.
                - "ROW_COUNT_IS_GREATER_THAN_OR_EQUAL_TO" -> The row count is greater than or equal to the threshold.
                - "ROW_COUNT_IS_WITHIN_A_RANGE" -> The row count is within the specified range.
                - "ROW_COUNT_GROWS_BY_AT_MOST_ABSOLUTE" -> The row count growth is at most the threshold (absolute change).
                - "ROW_COUNT_GROWS_BY_AT_LEAST_ABSOLUTE" -> The row count growth is at least the threshold (absolute change).
                - "ROW_COUNT_GROWS_WITHIN_A_RANGE_ABSOLUTE" -> The row count growth is within the specified range (absolute change).
                - "ROW_COUNT_GROWS_BY_AT_MOST_PERCENTAGE" -> The row count growth is at most the threshold (percentage change).
                - "ROW_COUNT_GROWS_BY_AT_LEAST_PERCENTAGE" -> The row count growth is at least the threshold (percentage change).
                - "ROW_COUNT_GROWS_WITHIN_A_RANGE_PERCENTAGE" -> The row count growth is within the specified range (percentage change).
                If not provided, the existing definition from the backend will be preserved (for update operations). Required when creating a new assertion (when urn is None).
            criteria_parameters (Optional[VolumeAssertionDefinitionParameters]): Optional threshold parameters to be used for the assertion. This can be a single threshold value or a tuple range.
                - If the condition is range-based (ROW_COUNT_IS_WITHIN_A_RANGE, ROW_COUNT_GROWS_WITHIN_A_RANGE_ABSOLUTE, ROW_COUNT_GROWS_WITHIN_A_RANGE_PERCENTAGE), the value is a tuple of two threshold values, with format (min, max).
                - For other conditions, the value is a single numeric threshold value.
                If not provided, existing value is preserved for updates. Required when creating a new assertion.
            time_bucketing_strategy (Optional[TimeBucketingStrategyInputTypes]): Optional time bucketing strategy for evaluating this assertion over time-based data partitions. Accepts a TimeBucketingStrategy Pydantic model, a dict, or a raw AssertionTimeBucketingStrategyClass. When updating, preserves the existing value if not provided. Bucket interval unit must be DAY or WEEK with multiple=1. Detection mechanism sourceType must be QUERY when bucketing is enabled.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            VolumeAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._volume_client.sync_volume_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            detection_mechanism=detection_mechanism,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
            criteria_condition=criteria_condition,
            criteria_parameters=criteria_parameters,
            time_bucketing_strategy=time_bucketing_strategy,
        )

    def sync_sql_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        statement: Optional[str] = None,
        criteria_condition: Optional[Union[SqlAssertionCondition, str]] = None,
        criteria_parameters: Optional[
            Union[Union[float, int], tuple[Union[float, int], Union[float, int]]]
        ] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        skip_dataset_exists_check: bool = False,
    ) -> SqlAssertion:
        """Upsert and merge a sql assertion.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schedule behavior:
            - Create case: Uses default daily schedule ("0 0 * * *") or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            statement (Optional[str]): The SQL statement to be used for the assertion. Required when creating a new assertion (urn=None), optional when updating an existing assertion.
            criteria_condition (Optional[Union[SqlAssertionCondition, str]]): The condition for the sql assertion. Required when creating a new assertion (urn=None), optional when updating an existing assertion. Valid values are:
                - "IS_EQUAL_TO" -> The metric value equals the threshold.
                - "IS_NOT_EQUAL_TO" -> The metric value does not equal the threshold.
                - "IS_GREATER_THAN" -> The metric value is greater than the threshold.
                - "IS_LESS_THAN" -> The metric value is less than the threshold.
                - "IS_WITHIN_A_RANGE" -> The metric value is within the specified range.
                - "GROWS_AT_MOST_ABSOLUTE" -> The metric growth is at most the threshold (absolute change).
                - "GROWS_AT_MOST_PERCENTAGE" -> The metric growth is at most the threshold (percentage change).
                - "GROWS_AT_LEAST_ABSOLUTE" -> The metric growth is at least the threshold (absolute change).
                - "GROWS_AT_LEAST_PERCENTAGE" -> The metric growth is at least the threshold (percentage change).
                - "GROWS_WITHIN_A_RANGE_ABSOLUTE" -> The metric growth is within the specified range (absolute change).
                - "GROWS_WITHIN_A_RANGE_PERCENTAGE" -> The metric growth is within the specified range (percentage change).
            criteria_parameters (Optional[Union[float, int, tuple[float, int]]]): The threshold parameters to be used for the assertion. Required when creating a new assertion (urn=None), optional when updating an existing assertion. This can be a single threshold value or a tuple range.
                - If the condition is range-based (IS_WITHIN_A_RANGE, GROWS_WITHIN_A_RANGE_ABSOLUTE, GROWS_WITHIN_A_RANGE_PERCENTAGE), the value is a tuple of two threshold values, with format (min, max).
                - For other conditions, the value is a single numeric threshold value.
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default daily schedule will be used. The format is a cron expression, e.g. "0 0 * * *" for daily at midnight using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SqlAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._sql_client.sync_sql_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            statement=statement,
            criteria_condition=criteria_condition,
            criteria_parameters=criteria_parameters,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
        )

    def sync_smart_sql_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        statement: Optional[str] = None,
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        sensitivity: Optional[Union[str, InferenceSensitivity]] = None,
        exclusion_windows: Optional[ExclusionWindowInputTypes] = None,
        training_data_lookback_days: Optional[int] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        skip_dataset_exists_check: bool = False,
    ) -> SmartSqlAssertion:
        """Upsert and merge a smart SQL assertion with AI-powered inference.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Smart SQL assertions use machine learning to infer appropriate thresholds for your
        SQL query results, rather than requiring you to specify fixed threshold values.

        Schedule behavior:
            - Create case: Uses default schedule of every 6 hours or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            statement (Optional[str]): The SQL statement to be used for the assertion. Required when creating a new assertion (urn=None), optional when updating an existing assertion.
                The SQL query should return a single numeric value. Example: "SELECT COUNT(*) FROM table WHERE status = 'active'"
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            sensitivity (Optional[Union[str, InferenceSensitivity]]): The sensitivity level for AI inference. Valid values are: "low", "medium", "high".
                - "low": Less sensitive, fewer alerts for anomalies
                - "medium": Balanced sensitivity (default)
                - "high": More sensitive, more alerts for smaller deviations
            exclusion_windows (Optional[ExclusionWindowInputTypes]): The exclusion windows to be applied to the assertion. Only fixed range exclusion windows are supported. Valid values are:
                - {"start": "2025-01-01T00:00:00", "end": "2025-01-02T00:00:00"} (using ISO strings)
                - {"start": datetime(2025, 1, 1, 0, 0, 0), "end": datetime(2025, 1, 2, 0, 0, 0)} (using datetime objects)
                - FixedRangeExclusionWindow(start=datetime(2025, 1, 1, 0, 0, 0), end=datetime(2025, 1, 2, 0, 0, 0)) (using typed object)
                - A list of any of the above formats
            training_data_lookback_days (Optional[int]): The number of days of historical data to use for training the AI model.
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default schedule of every 6 hours will be used. The format is a cron expression, e.g. "0 */6 * * *" for every 6 hours using UTC timezone. Alternatively, a models.CronScheduleClass object can be provided.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SmartSqlAssertion: The created or updated assertion.
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._smart_sql_client.sync_smart_sql_assertion(
            dataset_urn=dataset_urn,
            statement=statement,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            sensitivity=sensitivity,
            exclusion_windows=exclusion_windows,
            training_data_lookback_days=training_data_lookback_days,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
        )

    def sync_schema_assertion(
        self,
        *,
        dataset_urn: Union[str, DatasetUrn],
        urn: Optional[Union[str, AssertionUrn]] = None,
        display_name: Optional[str] = None,
        enabled: Optional[bool] = None,
        compatibility: Optional[Union[str, SchemaAssertionCompatibility]] = None,
        fields: Optional[SchemaAssertionFieldsInputType] = None,
        incident_behavior: Optional[AssertionIncidentBehaviorInputTypes] = None,
        tags: Optional[TagsInputType] = None,
        updated_by: Optional[Union[str, CorpUserUrn]] = None,
        schedule: Optional[Union[str, models.CronScheduleClass]] = None,
        skip_dataset_exists_check: bool = False,
    ) -> SchemaAssertion:
        """Upsert and merge a schema assertion to validate dataset schema.

        Note:
            Keyword arguments are required.

        Upsert and merge is a combination of create and update. If the assertion does not exist,
        it will be created. If it does exist, it will be updated. Existing assertion fields will
        be updated if the input value is not None. If the input value is None, the existing value
        will be preserved. If the input value can be un-set (e.g. by passing an empty list or
        empty string), it will be unset.

        Schema assertions validate that a dataset's schema matches expected field definitions
        with configurable compatibility modes (EXACT_MATCH, SUPERSET, SUBSET).

        Schedule behavior:
            - Create case: Uses default schedule of every 6 hours or provided schedule
            - Update case: Uses existing schedule or provided schedule.

        Args:
            dataset_urn (Union[str, DatasetUrn]): The urn of the dataset to be monitored.
            urn (Optional[Union[str, AssertionUrn]]): The urn of the assertion. If not provided, a urn will be generated and the assertion will be created in the DataHub instance.
            display_name (Optional[str]): The display name of the assertion. If not provided, a random display name will be generated.
            enabled (Optional[bool]): Whether the assertion is enabled. If not provided, the existing value will be preserved.
            compatibility (Optional[Union[str, SchemaAssertionCompatibility]]): The compatibility mode for schema validation. Valid values are:
                - "EXACT_MATCH" -> The schema must exactly match the expected fields (default).
                - "SUPERSET" -> The schema must contain at least all the expected fields.
                - "SUBSET" -> The schema must be a subset of the expected fields.
            fields (Optional[SchemaAssertionFieldsInputType]): The expected schema fields. Required when creating a new assertion (urn=None).
                Each field can be specified as:
                - A dict with 'path' and 'type' keys: {"path": "column_name", "type": "STRING"}
                - A dict with optional 'native_type': {"path": "id", "type": "NUMBER", "native_type": "BIGINT"}
                - A SchemaAssertionField object
                Valid type values are: BYTES, FIXED, BOOLEAN, STRING, NUMBER, DATE, TIME, ENUM, NULL, ARRAY, MAP, STRUCT, UNION
            incident_behavior (Optional[Union[str, list[str], AssertionIncidentBehavior, list[AssertionIncidentBehavior]]]): The incident behavior to be applied to the assertion. Valid values are: "raise_on_fail", "resolve_on_pass", or the typed ones (AssertionIncidentBehavior.RAISE_ON_FAIL and AssertionIncidentBehavior.RESOLVE_ON_PASS).
            tags (Optional[TagsInputType]): The tags to be applied to the assertion. Valid values are: a list of strings, TagUrn objects, or TagAssociationClass objects.
            updated_by (Optional[Union[str, CorpUserUrn]]): Optional urn of the user who updated the assertion. The format is "urn:li:corpuser:<username>". The default is the datahub system user.
            schedule (Optional[Union[str, models.CronScheduleClass]]): Optional cron formatted schedule for the assertion. If not provided, a default schedule of every 6 hours will be used. The format is a cron expression, e.g. "0 */6 * * *" for every 6 hours using UTC timezone.
            skip_dataset_exists_check (bool): If False (default), verifies the dataset_urn exists before creating/updating the assertion.
                Set to True when creating assertions before ingesting datasets (e.g., setting up assertions in a new environment
                before running ingestion pipelines), or when the dataset exists but may not be visible to the current API endpoint.

        Returns:
            SchemaAssertion: The created or updated assertion.

        Example:
            ```python
            # Create a schema assertion with exact match
            assertion = client.assertions.sync_schema_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,db.schema.table,PROD)",
                display_name="Expected Schema Check",
                compatibility="EXACT_MATCH",
                fields=[
                    {"path": "id", "type": "STRING"},
                    {"path": "count", "type": "NUMBER"},
                    {"path": "created_at", "type": "TIME"},
                ],
                enabled=True
            )

            # Create a superset schema assertion (actual schema must contain at least these fields)
            assertion = client.assertions.sync_schema_assertion(
                dataset_urn="urn:li:dataset:(urn:li:dataPlatform:snowflake,db.schema.table,PROD)",
                compatibility="SUPERSET",
                fields=[
                    {"path": "id", "type": "STRING"},
                    {"path": "name", "type": "STRING"},
                ],
            )
            ```
        """
        self._check_dataset_exists(dataset_urn, skip_dataset_exists_check)
        return self._schema_client.sync_schema_assertion(
            dataset_urn=dataset_urn,
            urn=urn,
            display_name=display_name,
            enabled=enabled,
            compatibility=compatibility,
            fields=fields,
            incident_behavior=incident_behavior,
            tags=tags,
            updated_by=updated_by,
            schedule=schedule,
        )



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import DataHubPageTemplateAssetSummaryClass
from .....schema_classes import DataHubPageTemplatePropertiesClass
from .....schema_classes import DataHubPageTemplateRowClass
from .....schema_classes import DataHubPageTemplateSurfaceClass
from .....schema_classes import DataHubPageTemplateVisibilityClass
from .....schema_classes import PageTemplateScopeClass
from .....schema_classes import PageTemplateSurfaceTypeClass
from .....schema_classes import SummaryElementClass
from .....schema_classes import SummaryElementTypeClass


DataHubPageTemplateAssetSummary = DataHubPageTemplateAssetSummaryClass
DataHubPageTemplateProperties = DataHubPageTemplatePropertiesClass
DataHubPageTemplateRow = DataHubPageTemplateRowClass
DataHubPageTemplateSurface = DataHubPageTemplateSurfaceClass
DataHubPageTemplateVisibility = DataHubPageTemplateVisibilityClass
PageTemplateScope = PageTemplateScopeClass
PageTemplateSurfaceType = PageTemplateSurfaceTypeClass
SummaryElement = SummaryElementClass
SummaryElementType = SummaryElementTypeClass

# fmt: on



"""
Constants specific to column metric assertions.

This module contains constants that are specific to column metric assertions.
Shared constants have been moved to column_assertion_constants.py.
"""

from enum import Enum
from typing import Union

# Re-export shared constants for backward compatibility
from acryl_datahub_cloud.sdk.assertion_input.column_assertion_constants import (  # noqa: F401
    ALLOWED_COLUMN_TYPES_FOR_COLUMN_ASSERTION,
    FIELD_VALUES_OPERATOR_CONFIG,
    OperatorInputType,
    OperatorType,
    RangeInputType,
    RangeTypeInputType,
    RangeTypeParsedType,
    ValueInputType,
    ValueType,
    ValueTypeInputType,
)
from datahub.metadata import schema_classes as models

# Backward compatibility alias
ALLOWED_COLUMN_TYPES_FOR_COLUMN_METRIC_ASSERTION = (
    ALLOWED_COLUMN_TYPES_FOR_COLUMN_ASSERTION
)

# Keep this in sync with FIELD_METRIC_TYPE_CONFIG in the frontend
# datahub-web-react/src/app/entityV2/shared/tabs/Dataset/Validations/assertion/builder/steps/field/utils.ts
FIELD_METRIC_TYPE_CONFIG = {
    "STRING": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
        models.FieldMetricTypeClass.MAX_LENGTH,
        models.FieldMetricTypeClass.MIN_LENGTH,
        models.FieldMetricTypeClass.EMPTY_COUNT,
        models.FieldMetricTypeClass.EMPTY_PERCENTAGE,
    ],
    "NUMBER": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
        models.FieldMetricTypeClass.MAX,
        models.FieldMetricTypeClass.MIN,
        models.FieldMetricTypeClass.MEAN,
        models.FieldMetricTypeClass.MEDIAN,
        models.FieldMetricTypeClass.STDDEV,
        models.FieldMetricTypeClass.NEGATIVE_COUNT,
        models.FieldMetricTypeClass.NEGATIVE_PERCENTAGE,
        models.FieldMetricTypeClass.ZERO_COUNT,
        models.FieldMetricTypeClass.ZERO_PERCENTAGE,
    ],
    "BOOLEAN": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "DATE": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "TIME": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
    "NULL": [
        models.FieldMetricTypeClass.NULL_COUNT,
        models.FieldMetricTypeClass.NULL_PERCENTAGE,
        models.FieldMetricTypeClass.UNIQUE_COUNT,
        models.FieldMetricTypeClass.UNIQUE_PERCENTAGE,
    ],
}


# Keep in sync with SMART_ASSERTION_ALLOWED_FIELD_METRIC_TYPES in frontend utils.ts
# and ALLOWED_SMART_FIELD_METRIC_TYPES in MonitorAssertionValidator.java
ALLOWED_SMART_FIELD_METRIC_TYPES: set[str] = {
    models.FieldMetricTypeClass.NULL_COUNT,
    models.FieldMetricTypeClass.UNIQUE_COUNT,
    models.FieldMetricTypeClass.EMPTY_COUNT,
    models.FieldMetricTypeClass.NEGATIVE_COUNT,
    models.FieldMetricTypeClass.ZERO_COUNT,
}


class MetricType(str, Enum):
    """Enum for field metric types used in column metric assertions."""

    NULL_COUNT = models.FieldMetricTypeClass.NULL_COUNT
    NULL_PERCENTAGE = models.FieldMetricTypeClass.NULL_PERCENTAGE
    UNIQUE_COUNT = models.FieldMetricTypeClass.UNIQUE_COUNT
    UNIQUE_PERCENTAGE = models.FieldMetricTypeClass.UNIQUE_PERCENTAGE
    MAX_LENGTH = models.FieldMetricTypeClass.MAX_LENGTH
    MIN_LENGTH = models.FieldMetricTypeClass.MIN_LENGTH
    EMPTY_COUNT = models.FieldMetricTypeClass.EMPTY_COUNT
    EMPTY_PERCENTAGE = models.FieldMetricTypeClass.EMPTY_PERCENTAGE
    MIN = models.FieldMetricTypeClass.MIN
    MAX = models.FieldMetricTypeClass.MAX
    MEAN = models.FieldMetricTypeClass.MEAN
    MEDIAN = models.FieldMetricTypeClass.MEDIAN
    STDDEV = models.FieldMetricTypeClass.STDDEV
    NEGATIVE_COUNT = models.FieldMetricTypeClass.NEGATIVE_COUNT
    NEGATIVE_PERCENTAGE = models.FieldMetricTypeClass.NEGATIVE_PERCENTAGE
    ZERO_COUNT = models.FieldMetricTypeClass.ZERO_COUNT
    ZERO_PERCENTAGE = models.FieldMetricTypeClass.ZERO_PERCENTAGE


# Type alias specific to column metric assertions
MetricInputType = Union[MetricType, models.FieldMetricTypeClass, str]
